import socket
from datetime import datetime
from urllib.parse import urlparse

from OpenSSL import SSL


async def get_translated_resource(hub, cert, url):
    """
    Converts the raw ssl certificate response to idem exec state output format
    Args:
        cert: TLS certificate details
        url(string): The URL of the website to get the certificates from.
    Returns: Dict with certificate details in output format
    """
    resource_translated = {"url": url}
    resource_translated["resource_id"] = cert.get_serial_number()
    resource_translated["sha1_fingerprint"] = (
        cert.digest("sha1").decode("utf-8").replace(":", "").lower()
    )
    resource_translated["issuer"] = {
        k.decode("utf8"): v.decode("utf8")
        for k, v in dict(cert.get_issuer().get_components()).items()
    }
    resource_translated["not_after"] = str(
        datetime.strptime(cert.get_notAfter().decode("utf-8"), "%Y%m%d%H%M%SZ")
    )
    resource_translated["not_before"] = str(
        datetime.strptime(cert.get_notBefore().decode("utf-8"), "%Y%m%d%H%M%SZ")
    )
    resource_translated["subject"] = {
        k.decode("utf8"): v.decode("utf8")
        for k, v in dict(cert.get_subject().get_components()).items()
    }
    resource_translated["version"] = str(cert.get_version())
    resource_translated["signature_algorithm"] = cert.get_signature_algorithm().decode(
        "utf-8"
    )
    resource_translated["serial_number"] = cert.get_serial_number()
    return resource_translated


async def get_ssl_connection(hub, ctx, url):
    """
    Get the SSL connection for given website URL
    Args:
        url(string): The URL of the website to get the certificates from.
    Returns: SSL connection for url
    """
    method = SSL.TLSv1_METHOD
    # Retrieve method from credentials if provided
    if ctx.acct.get("method", None) and getattr(SSL, f"{ctx.acct.method}_METHOD", None):
        method = getattr(SSL, f"{ctx.acct.method}_METHOD", None)
    url_parts = urlparse(url).netloc.split(":")
    host = url_parts[0].strip()
    port = int(url_parts[1]) if len(url_parts) > 1 else 443
    context = SSL.Context(method=method)
    conn = SSL.Connection(
        context, socket=socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    )
    conn.settimeout(5)
    conn.connect((host, port))
    conn.setblocking(1)
    return conn



from typing import Dict

from dict_tools.typing import Computed
from OpenSSL import SSL

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, url: str) -> Computed[Dict]:
    """
    Get information about the TLS certificates securing a host.


    Args:
        url(string):
            The URL of the website to get the certificates from.

    Request Syntax:
        [Idem-state-name]:
          exec.run:
            - path: tls.certificate.get
            - kwargs:
                url: 'string'

    Sample response:
            url: https://oidc.eks.us-east-2.amazonaws.com/id/sample
            sha1_fingerprint: 9e99a48a9960b14926bb7f3b02e22da2b0ab7280
            issuer:
              C: US
              O: Starfield Technologies, Inc.
              OU: Starfield Class 2 Certification Authority
            not_after: 2034-06-28 17:39:16
            not_before: 2009-09-02 00:00:00
            subject:
              C: US
              ST: Arizona
              L: Scottsdale
              O: Starfield Technologies, Inc.
              CN: Starfield Services Root Certificate Authority - G2
            version: 2
            signature_algorithm: sha256WithRSAEncryption
            serial_number: 12037640545166866303
            resource_id: 12037640545166866303

    Example Usage:
            unmanaged-tls_certificate:
              exec.run:
                - path: tls.certificate.get
                - kwargs:
                    url: https://oidc.eks.us-east-2.amazonaws.com/id/sample

            arn:aws:iam::537227425989:www.sample-2.com:
              aws.iam.open_id_connect_provider.absent:
                - name: www.sample-2.com
                - url: https://www.sample-2.com/
                - client_id_list:
                    - sts.amazonaws.com
                - thumbprint_list:
                    - ${exec:unmanaged-tls_certificate:sha1_fingerprint}
                - tags:
                    - Key: alpha.eksctl.io/eksctl-version
                      Value: 0.77.0
                    - Key: alpha.eksctl.io/cluster-name
                      Value: pr-ssc-eks-poc
    """
    result = dict(comment=[], ret=None, result=True)

    try:
        conn = await hub.tool.tls.certificate.get_ssl_connection(ctx, url)
        conn.do_handshake()
    except SSL.Error as e:
        # ignore handshake error, connection has server certificate chain
        hub.log.info(f"SSL handshake error {e.__class__.__name__}: {e}")
    except Exception as e:
        result["comment"].append(f"{e.__class__.__name__}: {e}")
        result["result"] = False
        return result

    try:
        cert = conn.get_peer_cert_chain()[-1]
    finally:
        conn.close()

    resource_translated = await hub.tool.tls.certificate.get_translated_resource(
        cert, url
    )

    result["comment"].append(f"fetched certificate information for url {url}")
    result["ret"] = resource_translated
    return result


async def list_(hub, ctx, url: str) -> Computed[Dict]:
    """
    Get list of TLS certificate in chain securing a host.

    List information about TLS certificates.

    Args:
        url(string):
            The URL of the website to get the certificates from.

    """
    result = dict(comment=[], ret=[], result=True)

    try:
        conn = await hub.tool.tls.certificate.get_ssl_connection(ctx, url)
        conn.do_handshake()
    except SSL.Error as e:
        # ignore handshake error, connection has server certificate chain
        hub.log.info(f"SSL handshake error {e.__class__.__name__}: {e}")
    except Exception as e:
        result["comment"].append(f"{e.__class__.__name__}: {e}")
        result["result"] = False
        return result

    try:
        certs = conn.get_peer_cert_chain()
    finally:
        conn.close()

    ret = []
    for cert in certs:
        resource_translated = await hub.tool.tls.certificate.get_translated_resource(
            cert, url
        )
        ret.append(resource_translated)
    result["ret"] = ret

    result["comment"].append(f"fetched certificates information for url {url}")
    return result



async def gather(hub, profiles):
    """
    load tls profiles from credential files

    Example:
    .. code-block:: yaml

        tls:
          default:
            method: TLSv1_2
    """
    sub_profiles = {}
    for profile, ctx in profiles.get("tls", {}).items():
        sub_profiles[profile] = ctx

    tls_profile = (hub.OPT.get("idem") or {}).get("acct_profile")
    if tls_profile not in sub_profiles:
        sub_profiles[tls_profile] = {"": ""}

    return sub_profiles
